/*
 * Author: wangjian
 * 生成ServiceImpl
 */
package com.tofly.xrtygis.service.impl;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.UpdateWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.tofly.common.core.util.DateUtil;
import com.tofly.common.core.util.StringUtil;
import com.tofly.common.oauth.exception.ToflyDeniedException;
import com.tofly.pojo.message.Messagepush;
import com.tofly.pojo.message.Messagerecipient;
import com.tofly.xrtygis.custom.entity.dma.ConstomTableList;
import com.tofly.xrtygis.custom.entity.enums.CompanyName;
import com.tofly.xrtygis.custom.entity.enums.GisMessage;
import com.tofly.xrtygis.custom.util.Date7Utils;
import com.tofly.xrtygis.custom.util.DateTimeUtil;
import com.tofly.xrtygis.em.Constant;
import com.tofly.xrtygis.em.ExcelExportFileTypeEnum;
import com.tofly.xrtygis.entity.Dmanightflow;
import com.tofly.xrtygis.entity.Dmatable;
import com.tofly.xrtygis.service.DmanightflowService;
import com.tofly.xrtygis.service.DmatableService;
import com.tofly.xrtygis.strategy.ExcelStrategyContext;
import com.tofly.xrtygis.util.ExcelUtil;
import com.xxl.job.core.biz.model.ReturnT;
import lombok.AllArgsConstructor;
import org.springframework.stereotype.Service;
import com.tofly.xrtygis.mapper.DmanightflowMapper;
import org.springframework.transaction.annotation.Transactional;

import javax.servlet.http.HttpServletResponse;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Collectors;

/**
 * 夜间最小流量分析表
 *
 * @author wangjieliu
 * @date Tue May 25 00:00:00 CST 2021
 */
@Service("dmanightflowService")
@AllArgsConstructor
public class DmanightflowServiceImpl extends ServiceImpl<DmanightflowMapper, Dmanightflow> implements DmanightflowService {


    private final ExcelStrategyContext excelStrategyContext;
    private final DmatableService dmatableService;

    @Override
    public IPage<Map> getMinFlow(HttpServletResponse response, String tableIds, String startTime, String endTime, String flag, String templateFlag,String isState, Page page) {
        //判断时间
        if (StringUtil.isEmpty(startTime) || StringUtil.isEmpty(endTime)) {
            throw new ToflyDeniedException("请选择月份时间");
        }
        if (StringUtil.isEmpty(tableIds)) {
            throw new ToflyDeniedException("请传入仪表ID");
        }
        //格式化时间
        Date startDate = DateUtil.formatDate(startTime, DateTimeUtil.FULL_FORMAT);
        Date endDate = DateUtil.formatDate(endTime, DateTimeUtil.FULL_FORMAT);
        //计算3个月后的时间
        Date addDate = Date7Utils.addNextPeriod(3, startDate);
        //比较时间是否大于3个月之后的时间
        if (endDate.after(addDate)) {
            //如果超过了3个月，则最多取到3个月的数据
            endDate = addDate;
        }
        if ("1".equals(flag)) {
            return baseMapper.getMinFlowCollet(tableIds, startDate, endDate,isState,page);
        } else if ("2".equals(flag)) {
            //需要导出的
            if ("1".equals(templateFlag)) {
                List<Map> mapList = baseMapper.getMinFlowDetail(tableIds, startDate, endDate,isState);
                //修改时间戳
//                mapList.forEach(k->{
//                    k.put("uday",k.get("uday").toString().substring(0,10));
//                });
                // 根据templateServiceType,templateFlag获取对应模板数据
                ExcelExportFileTypeEnum excelExportFileTypeEnum = ExcelExportFileTypeEnum.getExcelExportFileTypeEnum("nightFlowServiceImpl", templateFlag);
                if (null == excelExportFileTypeEnum) {
                    return null;
                }
                //将查询的结果转json字符串的格式，因为导出的时候只能传一个字符串；toJSONStringWithDateFormat可以转换时间格式字符串
                String resultListToString = JSONArray.toJSONStringWithDateFormat(mapList,"yyyy-MM-dd");
                // 从模型类型枚举根据templateServiceType获取策略对应模板数据
                List<Map<String, Object>> dataList = excelStrategyContext.getExportData("nightFlowServiceImpl", templateFlag, resultListToString);
                ExcelUtil.export(dataList.get(0), excelExportFileTypeEnum, Constant.FILE_PATH_TEST_EXPORT_EXCEL, response);
                return null;
            }else {
                IPage<Map> iPage = baseMapper.getMinFlowDetailPage(tableIds, startDate, endDate , page,isState);
                List<Map> mapList = iPage.getRecords();
                //修改时间戳
                mapList.forEach(k->{
                    k.put("uday",k.get("uday").toString().substring(0,10));
                });
                iPage.setRecords(mapList);
                return iPage;
            }
        }
        return null;
    }

    @Override
    public ConstomTableList getTableTree(String isFlow) {
        //查询所有区域观察表
        List<Dmatable> dmatableList = dmatableService.list(Wrappers.<Dmatable>lambdaQuery().eq("1".equals(isFlow) , Dmatable::getIsflow , "1")
                .eq("2".equals(isFlow) , Dmatable::getIspress , "1"));
        // 自定义一个存放顶级公司名称的对象
        ConstomTableList dmatable = new ConstomTableList();
        dmatable.setId(0L);
        dmatable.setSchName(CompanyName.CD_TY_SW.getName());
        List<Map> mapList = JSONObject.parseArray(JSONObject.toJSONString(dmatableList) , Map.class);
        //存放区域观察表
        dmatable.setTableList(mapList);
        return new ConstomTableList();
    }
    @Override
    @Transactional(rollbackFor = Exception.class)
    public Boolean nightMinimumFlowAnalysis(Date date){
        // 获取当前日期观察表(2:00 ~ 4:00)流量
        LocalDateTime now = DateTimeUtil.fromLong2LocalDateTime(date.getTime());
        Date startTime = DateTimeUtil.convertLDTToDate(now.plusHours(2));
        Date endTime = DateTimeUtil.convertLDTToDate(now.plusHours(4));
        List<Dmanightflow> flowList = baseMapper.getFlowList(startTime, endTime);
        List<Dmanightflow> pressureList = baseMapper.getPressureList(startTime, endTime);
        // 清空时间端范围内数据
        this.remove(new QueryWrapper<Dmanightflow>().lambda().between(Dmanightflow::getMintime,startTime,endTime));
        flowList.forEach(info->{
            // 保证 流量与压力 采集时间在 5分钟之内
            LocalDateTime time1 = DateTimeUtil.fromLong2LocalDateTime(info.getMintime().getTime());
            Long tableId = info.getTableId();
            pressureList.forEach(pressure->{
                Long pressureTableId = pressure.getTableId();
                LocalDateTime time2 = DateTimeUtil.fromLong2LocalDateTime(pressure.getMintime().getTime());
                LocalDateTime startTime1 = time1.minusMinutes(5);
                LocalDateTime endTime1 = time1.plusMinutes(5);
                if(tableId.equals(pressureTableId)){
                    if(startTime1.isBefore(time2) || time2.isBefore(endTime1)){
                        info.setMinpval(pressure.getMinpval());
                    }

                }

            });
            Date time = new Date();
            info.setUday(time);
            info.setUtime(time);
            info.setUmonth(DateTimeUtil.getLocalDate(time,DateTimeUtil.MONTHS_FORMAT));
        });
        if(flowList.size()>0){
            // 获取异常信息发送短信给所有人
            List<Dmanightflow> collect = flowList.stream().filter(i -> "1".equals(i.getUsfl())).collect(Collectors.toList());
            AtomicReference<Long> messagePushId = new AtomicReference<>(baseMapper.getPlanTempMaxId("TF_YWPN_MESSAGEPUSH_W"));
            AtomicReference<Long> messageRecipientId = new AtomicReference<>(baseMapper.getPlanTempMaxId("TF_YWPN_MESSAGERECIPIENT_W"));
            // 获取所有用户
            List<Long> userIdList = baseMapper.getUserIdList();
            List<Messagepush> messagePushList = new ArrayList<>();
            List<Messagerecipient> messageRecipientList = new ArrayList<>();
            collect.forEach(data->{
                // 消息组合
                String tableName = data.getTableName();
                Date mintime = data.getMintime();
                Double minval = data.getMinval();
                Double tval = data.getTval();
                String mess = tableName+"设备在"+DateTimeUtil.getLocalDate(mintime,DateTimeUtil.FULL_FORMAT)+",夜间最小流量(监测值为: "+minval+" m³/h,阈值: "+tval+"m³/h) 超标:"+(minval-tval)+"m³/h";
                userIdList.forEach(userId->{
                    Messagepush messagepush = new Messagepush();
                    Long message1 = messagePushId.get();
                    Long recipient1 = messageRecipientId.get();
                    // 消息主题/消息类型/消息表名/消息表中id/路由地址跳转
                    messagepush.setId(message1);
                    messagepush.setTopic(GisMessage.YJZXLL_ERROR.getTopic());
                    messagepush.setType(GisMessage.YJZXLL_ERROR.getType());
                    messagepush.setTableName(GisMessage.YJZXLL_ERROR.getTableName());
                    messagepush.setMessage(mess);
                    messagepush.setTableId(data.getId());
                    // 跳转到工单页面
                    messagepush.setRouteIp("");
                    messagepush.setRouteIpApp("");
                    messagepush.setCreateUser(userId);
                    messagepush.setCreateTime(DateTimeUtil.convertLDTToDate(now));
                    // 消息与接收人信息处理
                    Messagerecipient messagerecipient = new Messagerecipient();
                    messagerecipient.setId(recipient1);
                    messagerecipient.setMessageId(message1);
                    messagerecipient.setRecipient(userId);
                    messagerecipient.setFlag("0");
                    messagePushList.add(messagepush);
                    messageRecipientList.add(messagerecipient);
                    Long idAndSet = message1+1;
                    Long andSet = recipient1+1;
                    messagePushId.set(idAndSet);
                    messageRecipientId.set(andSet);
                });
            });

            if(messagePushList.size()>0){
                baseMapper.addMessageList(messagePushList);
                baseMapper.addMessageRecipientList(messageRecipientList);
            }

            return this.saveBatch(flowList);
        }else {
            return true;
        }

    }

}
