/*
 * Author: wangjian
 * 生成ServiceImpl
 */
package com.tofly.scada.service.impl;

import cn.hutool.Hutool;
import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.crypto.SecureUtil;
import cn.hutool.crypto.asymmetric.Sign;
import cn.hutool.crypto.asymmetric.SignAlgorithm;
import cn.hutool.crypto.digest.HMac;
import cn.hutool.crypto.digest.MD5;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.tofly.common.core.entity.ResultRespone;
import com.tofly.common.oauth.exception.ToflyDeniedException;
import com.tofly.scada.entity.Scada;
import com.tofly.scada.entity.ScadaLjFlow;
import com.tofly.scada.entity.ScadaSynHis;
import com.tofly.scada.mapper.ScadaLjFlowMapper;
import com.tofly.scada.mapper.ScadaSynHisMapper;
import com.tofly.scada.service.ScadaLjFlowService;
import com.tofly.scada.service.ScadaSynHisService;
import com.tofly.scada.util.DateTimeUtil;
import com.tofly.scada.util.HttpClientUtil;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;


/**
 * 第三方同步记录
 *
 * @author admin
 * @date Wed Sep 07 00:00:00 CST 2022
 */
@Service("scadaSynHisService")
@Slf4j
public class ScadaSynHisServiceImpl extends ServiceImpl<ScadaSynHisMapper, ScadaSynHis> implements ScadaSynHisService {
    @Value("${syn.fn}")
    private String fn;

    @Value("${syn.fc}")
    private String fc;

    @Value("${syn.fh}")
    private String fh;

    @Value("${syn.userId}")
    private String userId;

    @Value("${syn.pc}")
    private String pc;

    @Value("${syn.url}")
    private String url;

    @Value("${syn.md5Key}")
    private String md5Key;

    @Value("${syn.type}")
    private String type;


    @Override
    public List<String> getSynCodeList(String type) {
        return baseMapper.getSynCodeList(type);
    }

    @Scheduled(cron = "0 0 7,16 * * ? ")
    @Async
    public void automaticSynchronousThirdParty(){
        try {
            Calendar calYersterday = Calendar.getInstance();
            calYersterday.add(Calendar.DATE, -2);
            synchronousThirdParty(null, DateUtil.format(calYersterday.getTime(), "yyyy-MM-dd"), DateUtil.format(new Date(), "yyyy-MM-dd"));
        } catch (Exception e) {
            log.error(e.getMessage());
            e.printStackTrace();
        }
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public void synchronousThirdParty(List<String> codes, String startDate, String endDate) throws IOException {
        // 获取未同步监测数据
        List<Scada> scadaSynHisList = baseMapper.getScadaSynHisList(codes, startDate, endDate);
        if (CollUtil.isEmpty(scadaSynHisList)) {
            return;
        }
        List<Scada> scadaSynHisThirdList = scadaSynHisList.stream()
                .filter(s -> s != null && StringUtils.isNotBlank(s.getScadaTimeStr()) && s.getScadaTimeStr().equals("00"))
                .collect(Collectors.toList());
        // 封装第三方请求参数
        Map<String, Object> dataMap = new LinkedHashMap<>(12);
        dataMap.put("fn", fn);
        dataMap.put("fc", fc);
        dataMap.put("fh", fh);
        dataMap.put("userid", userId);
        dataMap.put("pc", pc);
        long time = System.currentTimeMillis();
        dataMap.put("time",time);
        // 签名生成 sign=md5(fn&fc&fh&userid&pc&time&md5key);
        String md5Str = "fn=" + fn + "&" + "fc=" + fc + "&" + "fh=" + fh + "&" + "userid=" + userId + "&" + "pc=" + pc + "&" + "time=" + time + "&" + "key=" + md5Key;
        log.info("第三方请求签名生成参数: {}", md5Str);
        String md5 = HttpClientUtil.TraditionMd5(md5Str);
        dataMap.put("sign", md5);
        log.info("第三方请求参数大小: {}", scadaSynHisThirdList.size());
        String result = null;
        int sizec = scadaSynHisThirdList.size();
        //向上取整 , 获得次数，但是次数不多，只有几次，所以可以使用for循环分段查询
        int countc = (int) Math.ceil((double) sizec / 1000d);
        for (int i = 0; i < countc; i++) {
            int startIndex = i * 1000;
            int endIndex = 1000 + i * 1000;
            List<Scada> subYhbhList;
            //判断是否截取的最后一次数据，如果是最后一次数据，则不能使用endIndex这个参数去截取了，有可能会存在endIndex参数值超过索引最大值的情况
            if (i < countc - 1) {
                //分段采取数据 0~999 ； 1000~1999 , 2000~2999 .....//这里subList方法无法截取最后一位，所以需要多写一位索引
                subYhbhList = scadaSynHisThirdList.subList(startIndex, endIndex);
            } else {
                //分段采取最后的数据
                subYhbhList = scadaSynHisThirdList.subList(startIndex, scadaSynHisThirdList.size());
            }
            dataMap.put("jsondata", JSON.toJSONString(subYhbhList));
            result = HttpClientUtil.doPostJson(url, new HashMap<>(12), JSON.toJSONString(dataMap));
        }

        JSONObject jsonObject = JSON.parseObject(result);
        log.info("第三方返回:{}",jsonObject);
        if (jsonObject == null) {
            throw new ToflyDeniedException("第三方无结果返回");
        }
        Integer status = (Integer) jsonObject.get("status");
        if(1==status){
            // 批量更新推送数据同步状态
            List<Long> collect = scadaSynHisList.stream().filter(Objects::nonNull).map(Scada::getId).collect(Collectors.toList());
            if(collect.size()>0){
                int size = collect.size();
                //向上取整 , 获得次数，但是次数不多，只有几次，所以可以使用for循环分段查询
                int count = (int) Math.ceil((double) size / 1000d);
                for (int i = 0; i < count; i++) {
                    int startIndex = i * 1000;
                    int endIndex = 1000 + i * 1000;
                    List<Long> subYhbhList;
                    //判断是否截取的最后一次数据，如果是最后一次数据，则不能使用endIndex这个参数去截取了，有可能会存在endIndex参数值超过索引最大值的情况
                    if (i < count - 1) {
                        //分段采取数据 0~999 ； 1000~1999 , 2000~2999 .....//这里subList方法无法截取最后一位，所以需要多写一位索引
                        subYhbhList = collect.subList(startIndex, endIndex);
                    } else {
                        //分段采取最后的数据
                        subYhbhList = collect.subList(startIndex, collect.size());
                    }
                    baseMapper.updateBatchSynFlag(subYhbhList);
                }
            }
        }else {
            // 记录异常,返回失败
            log.error("新天第三方同步异常:{}",jsonObject.toJSONString());
        }

    }

}
